<h3>Downloads</h3>
<p>Current version: <b>{{ release }}</b></p>
<p>
Development version: <br> &nbsp; &nbsp;
<a href="https://github.com/xraypy/XrayDB">
https://github.com/xraypy/XrayDB</a> <br>

<hr>
<p
